#
# EffectiveBilipschitz.py
#
# Goal - Give rigorous proofs of all the computer-assisted lemmas in
# EffectiveBilipschitz.pdf using interval arithmetic.
#
# To run all verifications type the following at the command line:
# sage -python EffectiveBilipschitz.py


from sage.all import pi, sqrt, sin, cos, arctan, \
                     exp, log, sinh, cosh, arcsinh, \
                     RIF, Rational


# sage casts rationals to RIF (for interval arithmetic computations)
# or to float (for graphing).  We'll use Rationals for all constants
# except PI where we will use a verified interval or a float.


from Traverse import traverse


# We'll need some version of pi and sqrt(2)

def get_pi(x):
    if x in RIF:
        return RIF(pi)
    assert type(x) is float
    return float(pi)

def get_sqrt_two(r):
    if r in RIF:
        return RIF(sqrt(2))
    assert type(r) is float
    return float(sqrt(2))


# Def 4.20 - Def:Hfunction - related to the eccentricity of an ellipse
eccen = Rational("33957/10000")

#
# Verifying Theorem 4.26 -- MaxTubeInjectivity
#

# From 4.27 - Eqn:MaxTubeInjecGeneral - related to chordal distance (for a tube boundary)
chord = Rational("1361/1000")

# According to a non-rigorous computation the lower bound in 4.27
# Eqn:MaxTubeInjectivity is linearly approximated by
# 1.12276771739404*x - 0.16039538819914867.  We round towards -\infty
# to obtain the simpler decimals which we then must check.

# Since S(R) is defined piecewise, it suffices to check both pieces.
# First we suppose that Z \geq 1/\sqrt{3}.

# IR = injectivity radius
def IR_up(interval):
    Z = interval[0]
    return chord * sqrt( 1 - cos( sqrt(1 - Z**2) ) ) * ( (1 + Z**2) / (1 - Z**2) ) * arcsinh ( Z * sqrt(1 - Z**2) / (1 + Z**2) )

def IR_down(interval):
    # From 4.28 Eqn:MaxTubeInjecLinear
    A = Rational("11227/10000")
    B = Rational( "1604/10000")
    Z = interval[0]
    return A * Z - B

# For plotting in sage:
# import EffectiveBilipschitz
# P = plot(lambda x: EffectiveBilipschitz.IR_up([x]), (x, 1/sqrt(3), 1))
# Q = plot(lambda x: EffectiveBilipschitz.IR_down([x]), (x, 1/sqrt(3), 1))
# (P + Q).show()

def init_box_IR():
    rs3 = RIF(1/sqrt(3))
    t = RIF(rs3.lower(), Rational("99995/100000"))
    # the upper bound is ok - see the end of Thm:MaxTubeInjecLinear
    return [t]

# Now we assume that Z lies in [0,  1/\sqrt{3}], That is

def IR_up_small_Z(interval):
    Z = interval[0]
    trt = RIF(2 * sqrt(2))
    sr = trt * arcsinh( 1 / trt )
    return chord * sqrt( 1 - cos( sqrt(1 - Z**2) ) ) * ( Z / sqrt(1 - Z**2) ) * sr

def init_box_IR_small_Z():
    rs3 = RIF(1/sqrt(3))
    t = RIF(0.0, rs3.upper())
    return [t]

def verify_MaxTubeInjectivity():
    print( "Verifying Theorem 4.26 -- Thm:MaxTubeInjectivity" )
    traverse(IR_up, IR_down, init_box_IR, max_depth = 50, report = 10000)
    traverse(IR_up_small_Z, IR_down, init_box_IR_small_Z, max_depth = 50, report = 1000)

# Code checked and run 2020-10-20:
#
# sage: import EffectiveBilipschitz
# sage: %time EffectiveBilipschitz.verify_MaxTubeInjectivity()
#
# ('boxes cleared', '111734', 'max depth achieved', '27')
# ('boxes cleared', '12', 'max depth achieved', '5')
# CPU times: user 1min 9s, sys: 1.25 s, total: 1min 10s
# Wall time: 1min 36s


#
# Verifying Lemma 5.15 -- Meyerhoff
#

# Here is the statement: For all tau \in [0, \pi] there is some m \in
# {1, 2, 3, ..., 8} so that:
#
# cosh(m * 0.0996) - cos(m * tau) \leq 0.34932
#
# That is,
#
# min { cosh(m * 0.0996) - cos(m * tau) | m in range(1,9) } \leq 0.34932

def meyerhoff_down(interval):
    max_len = Rational("996/10000")
    tau = interval[0]
    if tau in RIF:
        max_len = RIF(max_len)
    else:
        assert type(tau) is float
        max_len = float(max_len)
    return min( [(cosh(m*max_len) - cos(m*tau)) for m in range(1,9)] )

def meyerhoff_up(interval):
    K = Rational("34932/100000")
    tau = interval[0]
    if tau in RIF:
        return RIF(K)
    assert type(tau) is float
    return float(K)

# plot(lambda tau: f_down([tau]), (tau, 0, pi))

def init_box_meyerhoff():
    PI = RIF(pi)
    tau = RIF(0, PI)
    return [tau]

def verify_Meyerhoff():
    print( "Verifying Lemma 5.15 -- Lem:Meyerhoff" )
    traverse(meyerhoff_up, meyerhoff_down, init_box_meyerhoff, max_depth = 50, report = 1000)

# Code checked and run 2020-10-20:
#
# sage: import EffectiveBilipschitz
# sage: %time EffectiveBilipschitz.verify_Meyerhoff()
#
# ('boxes cleared', '47', 'max depth achieved', '13')
# CPU times: user 47.4 ms, sys: 8.57 ms, total: 55.9 ms
# Wall time: 62.7 ms


#
# Verifying Lemma 7.18 -- AreaBoundCapped
#


### Constants - note that some of these are reused later.


# Lem 5.15 - Lem:Meyerhoff - maximum length to ensure a tube of rad >
# 0.531, by Meyerhoff
meyer_max = Rational("996/10000")

# Thm 7.19 - Thm:ShortStaysShort - upper bound on length of curve we are tracking
ell_max = Rational("735/10000")

# The following is a hack to subdivide domain for ell, so that the
# argument of haze_inv stays small even with large error bounds on
# ell.
ell_mid = ell_max / 2

def m_max(ell):
    return meyer_max - Rational("352/1000")*ell

# Def 7.2 - Def:FDefine - This is pi*eccen, truncated in
# the correct direction.
F_add = Rational("10667/1000")

# Def 7.2 - Def:FDefine - This is eccen/2 times the multiplicative
# constant on ell from Prop 6.15, truncated in the correct direction
F_mul = Rational("20977/1000")

# The error in the paper is given in the last line of 7.18.
ERROR = Rational("1/100000")

# However it is immediately multiplied by 2*pi.  We can always make
# the error smaller, so we define
PI_ERROR = Rational("62/10") * ERROR

### Functions

# Rem 4.23 - Rem:Haze - haze is is the [HK] h function expressed in
# terms of z = tanh R instead of R - haze means "h of z"
def haze(z):
    return eccen * ( z * (1 - z**2) / (1 + z**2) )

# Remark - haze likes numbers between 0.485+ and 1-

# plot( lambda z: EffectiveBilipschitz.haze(z), (z, 0.4851, 0.999) )

# See Eqn:HazeInv

# We now compute the inverse of haze using Cardano's formula.  Note
# that there are three branches of the inverse.  We chose the correct
# branch by checking it to high precision at a single (non-branch)
# point.

# Reference -
# "Viete, Descartes and the cubic equation" by Nickalls
# https://en.wikipedia.org/wiki/Cubic_function#Trigonometric_solution_for_three_real_roots

def haze_inv(h): # Again, see Eqn:HazeInv
    PI = get_pi(h)
    x = h / eccen
    return Rational("2/3") * sqrt( x**2 + 3 ) * cos( PI/3 + Rational("1/3") * arctan (-3 * sqrt( -3*x**4 - 33*x**2 + 3 ) / (x**3 + 18*x) ) ) - x/3

# haze_inv likes numbers between 0+ and H_max == 1.01967 (see
# immediately below Lemma 5.13)

# plot( lambda h: EffectiveBilipschitz.haze_inv(h), (h, 0.0, 1.01967) )

# One problem is that haze_inv crashes when fed a number that is too large,
# or a box whose upper bound is too large, or a box whose lower bound is too small.

# Now come functions related to Lemma 7.19.

# Both defined in Lem:AreaBoundCapped - Z_zero = Z_0
def Z_zero(ell):
    PI = get_pi(ell)
    return haze_inv(2*PI*ell)

def Z_min(ell, m):
    PI = get_pi(ell)
    return haze_inv(2*PI*(ell + m + ERROR))

# Defined in Lem:VisualAreaSigma
def q(z):
    return 1 + (3*z**2 - 1) / (z**2 * (3 - z**2) )

def q_zero(ell):
    return q(Z_zero(ell))

# Defined in Def:FDefine along with constants F_add and F_mul
def F(z, ell):
    return ( (1 + z**2) / (z**3 * (3 - z**2)) ) * ( ell / ( F_add - (F_mul * ell) ) )

# We divide by ell by hand to slightly simplify the evaluation.
def F_over_ell(z, ell):
    return ( (1 + z**2) / (z**3 * (3 - z**2)) ) * ( Rational("1") / (F_add - (F_mul * ell) ) )


# See the second to last paragraph of the proof of AreaBoundCapped for
# the usage of the next two functions
def Z_min_up(ell):
    return Z_min(ell, m_max(ell))

def F_up(ell):
    return F(Z_min_up(ell), ell)

# Here is the tex original for the function that we must check:

# flm(t) = f_{\ell,m} (t) =  2\pi \ell \left( \frac{\sqrt{t}}{2\pi} \right)**{q(Z_0)} +  2\pi m \exp((4\pi**2 - t) F(\Zmin, \ell) )

# As discussed in the second to last paragraph of the proof we plug in
# m_max(ell) for m.

def flm_up(t, ell):
    PI = get_pi(t)
    return 2*PI*ell * (sqrt(t) / (2*PI))**(q_zero(ell)) + 2*PI*m_max(ell) * exp((4*(PI)**2 - t) * F_up(ell))

# We prefer to check almost positivity

def glm_up(rect):
    t, ell = rect
    PI = get_pi(t)
    return 2*PI*(ell + m_max(ell)) - flm_up(t, ell)

def glm_down(rect):
    t, ell = rect
    return t.parent(-PI_ERROR)

# Here is the plot
# plot3d(g_up([t, ell]), (t,0.000001,4*pi**2), (ell, 0.0, ell_max))

# Subdivide domain into two subdomains, in order to prevent crashes in
# call to haze_inv

# First rectangle for g_up

def init_box_glm_small_ell():
    PI = RIF(pi)
    t = RIF(0.0, 4*PI**2) # this is slow near 0.0
    ell = RIF(0, ell_mid)
    return [t, ell]

# Second rectangle for g_up

def init_box_glm_big_ell():
    PI = RIF(pi)
    t = RIF(0.0, 4*PI**2) # this is slow near 0.0
    ell = RIF(ell_mid, ell_max)
    return [t, ell]

def verify_AreaBoundCapped():
    print( "Verifying Lemma 7.18 -- Lem:AreaBoundCapped" )
    traverse(glm_up, glm_down, init_box_glm_small_ell, max_depth = 60, report = 10000)
    traverse(glm_up, glm_down, init_box_glm_big_ell, max_depth = 50, report = 10000)

# Code checked and run 2020-10-20:
# %time EffectiveBilipschitz.verify_AreaBoundCapped()
# ('boxes cleared', '4573272', 'max depth achieved', '50')
# ('boxes cleared', '111474', 'max depth achieved', '30')
# CPU times: user 2h 26s, sys: 2min 36s, total: 2h 3min 2s


#
#  Verifying Lemma 7.26 - Lem:SFunctionProps
#

# Constants

# From Definition 7.25 Def:SystoleL
LenAdd = Rational("1603/100")
# LenAdd = 16.03

# From Lemma 7.26 Lem:SFunctionProps
ADD_SLOP = 58

# Functions

# From Definition 7.25 Def:SystoleL

def lmax(L):
    PI = get_pi(L)
    return 2*PI / (L**2 - LenAdd)

# From Equation 7.27 Eqn:SUpper
def sfun_up(interval):
    L = interval[0]
    PI = get_pi(L)
    return log( (L**2 - LenAdd) / (L**2 - ADD_SLOP) ) * (L**2 - LenAdd) / (2*PI)

def sfun_down(interval):
    L = interval[0]
    PI = get_pi(L)
    lm = lmax(L)
    return 4*PI**2 * F_over_ell( haze_inv(4*PI*lm + PI_ERROR), lm )

# P = plot(lambda L: sfun_up([L]), (L, 10.1, 11.0))
# Q = plot(lambda L: sfun_down([L]), (L, 10.1, 11.0))
# (P + Q).show()

def init_box_sfun():
    t = RIF(10.1, 11.0)
    return [t]


def verify_SFunctionProps():
    print( "Verifying Lemma 7.26 -- Lem:SFunctionProps" )
    traverse(sfun_up, sfun_down, init_box_sfun, max_depth = 50, report = 1000)


# Code checked and run on 2020-11-05:
# sage: %time EffectiveBilipschitz.verify_sfun()
# ('[11.?]', '100', '1000')
# [(10.0999999999999, 11.0000000000000)]
# ('boxes cleared', '22', 'max depth achieved', '10')
# CPU times: user 23.1 ms, sys: 77.1 ms, total: 100 ms
# Wall time: 276 ms


#
# Verifying Theorem 7.31, Thm:HoldShortGeodesics
#

# Maximum value of l+2m, from statement of Theorem
sum_max =  Rational("14 / 100")

# Functions
def two_m_max_hat(ell):
    return sum_max - ell

def Z_min_hat(ell):
    PI = get_pi(ell)
    return haze_inv(2*PI*(sum_max + ERROR))

def fhat_up(t, ell):
    PI = get_pi(t)
    return 2*PI*ell * (sqrt(t) / (2*PI))**(q_zero(ell)) + 2*PI*two_m_max_hat(ell) * exp((4*(PI)**2 - t) * F(Z_min_hat(ell), ell))

# We prefer to check almost positivity

def ghat_up(rect):
    t, ell = rect
    PI = get_pi(t)
    return 2*PI*sum_max - fhat_up(t, ell)

def ghat_down(rect):
    t, ell = rect
    return t.parent(-PI_ERROR)

# Subdivide domain into two subdomains, in order to
# prevent sqrt(neg) in call to haze_inv

def init_box_ghat_small_ell():
    PI = RIF(pi)
    t = RIF(0.0, 4*PI**2) # this is slow near 0.0
    ell = RIF(0.0, ell_mid)
    return [t, ell]

def init_box_ghat_big_ell():
    PI = RIF(pi)
    t = RIF(0.0, 4*PI**2) # this is slow near 0.0
    ell = RIF(ell_mid, ell_max)
    return [t, ell]


def verify_HoldShortGeodesics():
    print( "Verifying Theorem 7.31 -- Thm:HoldShortGeodesics" )
    traverse(ghat_up, ghat_down, init_box_ghat_small_ell, max_depth = 60, report = 50000)
    traverse(ghat_up, ghat_down, init_box_ghat_big_ell, max_depth = 60, report = 10000)


# Verified on 2020-11-13
# Verifying Theorem 7.31 -- Thm:HoldShortGeodesics
# ('[1.?e2, 0.1?]', '60', '50000')
# [(0.000000000000000, 39.4784176043575), (0.000000000000000, 0.0367500000000001)]
# ('boxes cleared', '17390252', 'max depth achieved', '50')
# ('[1.?e2, 0.1?]', '60', '10000')
# [(0.000000000000000, 39.4784176043575), (0.0367499999999999, 0.0735000000000001)]
# ('boxes cleared', '90830', 'max depth achieved', '27')


#
# Verifying Lemma 8.10, Lem:Puiseux
#

# Lem:Puiseux and Eqn:MeanValueMultiplier

# r is the radius, s is the square root of two, and m is a multiplicative constant,
# called C in the text.
# var('r, s, m')

# From Bromberg - Thm:MeanValue, making sqrt(2) the variable s.
# f(r, s) = cosh(r) * sin(s*r) - s * sinh(r) * cos(s*r)


# Sage commands for computing the tenth derivative of Phi, divided by 10 factorial.

# Phi(r, s, m) = (2 * (f(r, s))^2 * m^2 - 3 * r^3 * (sinh(2*r) - 2*r))

# DiffPhi(r, s, m) = Phi(r, s, m).derivative(r, 10)/factorial(10)

# DiffPhi(r, s, m) = DiffPhi(r, s, m).full_simplify()

# DiffPhi(r, s, m) = -4/14175*m^2*s^12*sinh(r)^2 + 2/14175*(52*m^2*sinh(r)^2 + 27*m^2)*s^10 - 2/945*(2*m^2*sinh(r)^2 + m^2)*s^8 - 8/675*(2*m^2*sinh(r)^2 + m^2)*s^6 - 2/945*(2*m^2*sinh(r)^2 + m^2)*s^4 + 2/525*(2*m^2*sinh(r)^2 + m^2)*s^2 - 64/14175*(m^2*s^11*cosh(r)*sinh(r) - 5*m^2*s^9*cosh(r)*sinh(r) - 6*m^2*s^7*cosh(r)*sinh(r) + 6*m^2*s^5*cosh(r)*sinh(r) + 5*m^2*s^3*cosh(r)*sinh(r) - m^2*s*cosh(r)*sinh(r))*cos(r*s)*sin(r*s) + 4/14175*(2*m^2*s^12*sinh(r)^2 - (52*m^2*sinh(r)^2 + 27*m^2)*s^10 + 15*(2*m^2*sinh(r)^2 + m^2)*s^8 + 84*(2*m^2*sinh(r)^2 + m^2)*s^6 + 15*(2*m^2*sinh(r)^2 + m^2)*s^4 + 2*m^2*sinh(r)^2 - 26*(2*m^2*sinh(r)^2 + m^2)*s^2 + m^2)*sin(r*s)^2 - 4/4725*(2*r^3 + 135*r)*cosh(r)*sinh(r) - 8/315*(r^2 + 6)*sinh(r)^2 - 4/315*r^2 - 8/105

# DiffPhiSub(r, m) = DiffPhi(r, sqrt(2), m)

# So here is the tenth derivative, divided by (10!) :
def DiffPhiSub(r, s, m):
    return Rational("64/225")*s*(m**2)*cos(s*r)*cosh(r)*sin(s*r)*sinh(r) -\
        Rational("8/189")*(m**2)*sinh(r)**2 +\
        Rational("4/4725")*(102*(m**2)*sinh(r)**2 + 19*(m**2))*sin(s*r)**2 -\
        Rational("4/4725")*(2*r**3 + 135*r)*cosh(r)*sinh(r) -\
        Rational("8/315")*(r**2 + 6)*sinh(r)**2 -\
        Rational("4/525")*m**2 - Rational("4/315")*r**2 - Rational("8/105")

# Constants

# Bounds on radius
# rmin = Rational("0")
# rmaxone = Rational("469/1000")
# rmaxtwo = Rational("53/1000")

# Multiplicative constant, called C in the text
# multone = 1.046
# multtwo = 1.001

def derfunone_up(interval):
    r = interval[0]
    s2 = get_sqrt_two(r)
    C = Rational("1046/1000")
    return DiffPhiSub(r, s2, C)

def derfuntwo_up(interval):
    r = interval[0]
    s2 = get_sqrt_two(r)
    C = Rational("1001/1000")
    return DiffPhiSub(r, s2, C)

def derfunone_down(interval):
    return Rational("-85/1000")

def derfuntwo_down(interval):
    return Rational("-84/1000")

# P = plot(lambda r: derfunone_up([r]), (r, rmin, rmaxone))
# Q = plot(lambda r: derfunone_down([r]), (r, rmin, rmaxone))
# (P + Q).show()

def init_box_derfunone():
    r = RIF(0.0, 0.469)
    return [r]

def init_box_derfuntwo():
    r = RIF(0.0, 0.053)
    return [r]

def verify_Puiseux():
    print( "Verifying Lemma 8.10 -- Lem:Puiseux" )
    traverse(derfunone_up, derfunone_down, init_box_derfunone, max_depth = 50, report = 1000)
    traverse(derfuntwo_up, derfuntwo_down, init_box_derfuntwo, max_depth = 50, report = 1000)

# sage: %time EffectiveBilipschitz.verify_Puiseux()
# ('[1.?]', '100', '1000')
# [(0.000000000000000, 0.469000000000000)]
# ('boxes cleared', '12', 'max depth achieved', '5')
# ('[0.1?]', '100', '1000')
# [(0.000000000000000, 0.0530000000000000)]
# ('boxes cleared', '4', 'max depth achieved', '2')
# CPU times: user 17.2 ms, sys: 3.52 ms, total: 20.7 ms
# Wall time: 18.6 ms


#
# Verifying Lemma 9.24 -- gJBehavior.
#

# Multiplicative constant in Theorem 4.15, comes from bilipschitz Theorem 8.18
ThickMult = Rational("4961/10")

# Additive constant in Theorem 4.15, comes from distance in tubes (Theorem 3.6)
ThickAdd = Rational("1475/10000")
half = Rational("1/2")

# From equation (9.17) and the statement of Lemma 9.24, gJBehavior
def gJ_down(rect):
    epsilon, J  = rect
    numerator = (epsilon**5)*log(J)
    denom = ThickMult * (J**5) * cosh(J*epsilon*half + ThickAdd)**5
    return numerator/denom

# From statement of Lemma 9.24, gJBehavior
def gJ_up(rect):
    epsilon, J = rect
    Bound = Rational("561/10000000")
    if J in RIF:
        return RIF(Bound)
    assert type(J) is float
    return float(Bound)

# Box is from statement of Lemma 9.24, gJBehavior, combined with the fact (established
# during the proof) that it suffices to let epsilon = log 3.
def init_box_gJ():
    eroot = RIF(exp(0.2))
    epsilon = RIF(log(3))
    J = RIF(1.0, eroot.upper())
    return [epsilon, J]

# Second term in factorization of \partial g / \partial J, in Lemma 9.24, gJBehavior
def gDerivTerm_up(rect):
    epsilon, J = rect
    twofifths = Rational("2/5")
    half = Rational("1/2")
    return (twofifths - 2*log(J))*cosh(J*epsilon*half + ThickAdd) - epsilon*log(J)*J* sinh(J*epsilon*half + ThickAdd)

def gDerivTerm_down(rect):
    epsilon, J = rect
    if J in RIF:
        return RIF(0)
    assert type(J) is float
    return float(0)

def init_box_gDeriv():
    logthree = RIF(log(3))
    epsilon = RIF(0.0, logthree.upper())
    J = RIF(1.0, 1.15)
    return [epsilon, J]


def verify_gJBehavior():
    print( "Verifying Lemma 9.24 -- Lem:gJBehavior" )
    traverse(gJ_up, gJ_down, init_box_gJ, max_depth = 50, report = 1000)
    traverse(gDerivTerm_up, gDerivTerm_down, init_box_gDeriv, max_depth = 50, report = 1000)


# Verified on 2022-07-22
# Verifying Lemma 9.24 -- Lem:gJBehavior
# ('[1.098612288668110?, 2.?]', '50', '1000')
# [(1.09861228866810, 1.09861228866811), (1.00000000000000, 1.22140275816017)]
# ('cleared', '1000', 'curr box', '[(1.09861228866810, 1.09861228866811), (1.14949821249548, 1.14951172584742)]', 'at depth', '14', 'left', '9')
# ('boxes cleared', '1158', 'max depth achieved', '16')
# ('[1.?, 1.1?]', '50', '1000')
# [(0.000000000000000, 1.09861228866811), (1.00000000000000, 1.15000000000000)]
# boxes cleared 7 max depth achieved 6
% CPU times: user 368 ms, sys: 16.4 ms, total: 385 ms
% Wall time: 405 ms

def verify_all():
    verify_MaxTubeInjectivity()
    verify_Meyerhoff()
    verify_AreaBoundCapped()
    verify_SFunctionProps()
    verify_HoldShortGeodesics()
    verify_Puiseux()
    verify_gJBehavior()

if __name__ == "__main__":
    verify_all()

# To run all verifications type the following at the command line:
# sage -python EffectiveBilipschitz.py
