#
# Traverse.py
#


# Goal - we provide a "tree traversal" method to prove, using interval
# arithmetic (RIF in Sage), an inequality for all points in a
# rectangular domain in \RR^d.


# Import

from copy import deepcopy
import gc


# Enable garbage collection to clean out memory. To turn on this option,
# set the COLLECT_GARBAGE variable to True.

gc.enable()
COLLECT_GARBAGE = False


# Splitting boxes

def split(box):
    '''
    Given a box, returns the two boxes obtained by bisecting the
    widest dimension.
    '''
    dim = len(box)
    max_dim = 0
    max_diam = 0
    for curr_dim in range(dim):
        curr_diam = box[curr_dim].diameter()
        if curr_diam > max_diam:
            max_dim = curr_dim
            max_diam = curr_diam

    # We need to return two boxes, so make deep copies.
    u0, u1 = box[max_dim].bisection()
    box0 = [deepcopy(x) for x in box]
    box1 = [deepcopy(x) for x in box]
    box0[max_dim] = u0
    box1[max_dim] = u1
    return box0, box1


# Traversing the tree

def traverse(up_func, down_func, init_box_func, max_depth = 100, report = 1000):
    init_box = init_box_func()
    print( str(init_box), str(max_depth), str(report) )
    print( str([a.endpoints() for a in init_box_func()]) )

    max_depth_achieved = 0
    todo = [(init_box, 0)]
    num_cleared = 0
    while todo != []:
        curr_box, depth = todo.pop()
        if depth > max_depth_achieved:
            max_depth_achieved = depth
        if up_func(curr_box) > down_func(curr_box):
            # all is well
            num_cleared += 1
            if COLLECT_GARBAGE and num_cleared % 2000 == 0:
                gc.collect() 
            if num_cleared % report == 0:
                curr_box_data = [x.endpoints() for x in curr_box]
                print( "cleared", str(num_cleared), "curr box", str(curr_box_data),
                       "at depth", str(depth), "left", str(len(todo)) )
            continue
        elif depth > max_depth:
            # report and give up
            curr_box_data = [x.endpoints() for x in curr_box]
            print( "cleared", str(num_cleared), "curr box", str(curr_box_data),
                   "exceeded max_depth", str(depth), "left", str(len(todo)) )
            print( "Run failed." )
            break
        else:
            # descend
            box0, box1 = split(curr_box)
            todo.append((box0, depth + 1))
            todo.append((box1, depth + 1))

    print( "boxes cleared", str(num_cleared),
           "max depth achieved", str(max_depth_achieved) )
